#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        plugin.add_option(
            "cmd_prefix",
            type="textarea",
            value="Execute voice command",
            label="Magic prefix for voice commands",
            description="Optional magic prefix required for voice commands, e.g. 'OK PyGPT', 'Execute voice command', etc.",
            urls={
                "Help": "https://pygpt.readthedocs.io/en/latest/accessibility.html",
            },
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import os.path
from typing import Optional

from llama_index.core import StorageContext, load_index_from_storage
from llama_index.core.indices.base import BaseIndex
from llama_index.core.indices.service_context import ServiceContext

from .base import BaseStore


class CtxAttachmentProvider(BaseStore):
    def __init__(self, *args, **kwargs):
        super(CtxAttachmentProvider, self).__init__(*args, **kwargs)
        """
        Context attachments vector store provider

        :param args: args
        :param kwargs: kwargs
        """
        self.window = kwargs.get('window', None)
        self.path = kwargs.get("path", None)
        self.id = "CtxAttachmentVectorStore"
        self.index = None

    def get_path(self, id: str) -> str:
        """
        Get database path

        :return: database path
        """
        return self.path

    def exists(
            self,
            id: Optional[str] = None
    ) -> bool:
        """
        Check if index exists

        :return: True if exists
        """
        path = self.get_path()
        if os.path.exists(path):
            store = os.path.join(path, "docstore.json")
            if os.path.exists(store):
                return True
        return False

    def create(self, id: str):
        """
        Create empty index

        :param id: index name (not used)
        """
        path = self.get_path(id)
        if not os.path.exists(path):
            index = self.index_from_empty()  # create empty index
            self.store(
                id=id,
                index=index,
            )
        else:
            self.index = self.index_from_empty()

    def get(
            self,
            id: str,
            service_context: Optional[ServiceContext] = None
    ) -> BaseIndex:
        """
        Get index

        :param id: index name (not used)
        :param service_context: Service context
        :return: index instance
        """
        if not self.exists():
            self.create(id)
        path = self.get_path(id)
        storage_context = StorageContext.from_defaults(
            persist_dir=path,
        )
        self.index = load_index_from_storage(
            storage_context,
            service_context=service_context,
        )

        return self.index

    def store(
            self,
            id: str,
            index: Optional[BaseIndex] = None
    ):
        """
        Store index

        :param id: index name (not used)
        :param index: index instance
        """
        path = self.get_path(id)
        index.storage_context.persist(
            persist_dir=path,
        )
        self.index = index

    def clean(self):
        """Clean index"""
        self.index = None
        path = self.get_path("")
        if os.path.exists(path):
            os.remove(path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import copy
from typing import Optional

from PySide6.QtWidgets import QApplication

from pygpt_net.item.assistant import AssistantItem
from pygpt_net.utils import trans


class Editor:
    def __init__(self, window=None):
        """
        Assistants editor controller

        :param window: Window instance
        """
        self.window = window
        self.options = {
            "id": {
                "type": "text",
                "label": "assistant.id",
                "read_only": True,
            },
            "name": {
                "type": "text",
                "label": "assistant.name",
            },
            "description": {
                "type": "text",
                "label": "assistant.description",
            },
            "model": {
                "type": "combo",
                "use": "models",
                "keys": [],
                "label": "assistant.model",
            },
            "instructions": {
                "type": "textarea",
                "label": "assistant.instructions",
            },
            "vector_store": {
                "type": "combo",
                "label": "assistant.vector_store",
            },
            "tool.code_interpreter": {
                "type": "bool",
                "label": "assistant.tool.code_interpreter",
                "value": True,
            },
            "tool.file_search": {
                "type": "bool",
                "label": "assistant.tool.file_search",
                "value": False,
            },
            "tool.function": {
                "type": "dict",
                "label": "assistant.tool.function",
                "keys": {
                    'name': 'text',
                    'params': 'textarea',
                    'desc': 'textarea',
                },
            },
        }
        self.id = "assistant"
        self.current = None

    def get_options(self) -> dict:
        """
        Get options list

        :return: options dict
        """
        return self.options

    def get_option(self, key: str) -> dict:
        """
        Get option by key

        :param key: option key
        :return: option dict
        """
        if key in self.options:
            return self.options[key]

    def setup(self):
        """Setup editor"""
        parent = "assistant"
        key = "tool.function"
        self.window.ui.dialogs.register_dictionary(
            key,
            parent,
            self.get_option(key),
        )
        self.update_store_list()  # vector store list

    def update_store_list(self):
        """Update vector store list"""
        items = {}
        items["--"] = "--"  # none
        current_idx = 0
        current_id = self.get_selected_store_id()
        if current_id is not None:
            current_idx = self.get_choice_idx_by_id(current_id)
        stores = self.window.core.assistants.store.get_all()
        for id in list(stores.keys()):
            if self.window.core.assistants.store.is_hidden(id):
                continue
            if stores[id].name is None or stores[id].name == "":
                items[id] = id
            else:
                items[id] = stores[id].name
        self.window.controller.config.update_combo(self.id, "vector_store", items)
        if current_id is not None:
            current_idx = self.get_choice_idx_by_id(current_id)
        option = copy.deepcopy(self.options["vector_store"])
        option["idx"] = current_idx
        self.window.controller.config.apply_value(
            parent_id=self.id,
            key="vector_store",
            option=option,
            value="",
        )

    def get_selected_store_id(self) -> Optional[str]:
        """
        Return current selected vector store ID

        :return: vector store ID or None
        """
        idx = self.window.controller.config.get_value(
            parent_id=self.id,
            key='vector_store',
            option=self.options['vector_store'],
            idx=True,  # return idx, not the text value
        )  # empty or not
        if idx > 0:
            return self.window.ui.config[self.id]['vector_store'].combo.itemData(idx)
        return None

    def get_choice_idx_by_id(self, store_id: str) -> int:
        """
        Return combo choice index by vector store ID

        :param store_id: store ID
        :return: combo idx
        """
        stores = self.window.core.assistants.store.get_all()
        i = 1
        for id in list(stores.keys()):
            if self.window.core.assistants.store.is_hidden(id):
                continue  # ignore empty names
            if id == store_id:
                return i
            i += 1
        return 0  # none

    def edit(self, idx: Optional[int] = None):
        """
        Open assistant editor

        :param idx: assistant index (row index)
        """
        id = None
        if idx is not None:
            id = self.window.core.assistants.get_by_idx(idx)

        self.init(id)
        self.window.ui.dialogs.open_editor('editor.assistants', idx, width=900)

    def close(self):
        """Close assistant editor"""
        self.window.ui.dialogs.close('editor.assistants')

    def init(self, id: Optional[str] = None):
        """
        Initialize assistant editor

        :param id: assistant ID (in API)
        """
        assistant = self.window.core.assistants.create()
        assistant.model = "gpt-4-1106-preview"  # default model

        # if editing existing assistant
        if id is not None and id != "":
            if self.window.core.assistants.has(id):
                assistant = self.window.core.assistants.get_by_id(id)
        else:
            # defaults
            assistant.instructions = 'You are a helpful assistant.'
            assistant.tools['code_interpreter'] = True
            assistant.tools['file_search'] = True

        self.update_store_list()

        self.current = assistant

        if assistant.name is None:
            assistant.name = ""
        if assistant.description is None:
            assistant.description = ""
        if assistant.instructions is None:
            assistant.instructions = ""
        if assistant.model is None:
            assistant.model = ""

        options = {}
        data_dict = assistant.to_dict()
        for key in self.options:
            options[key] = copy.deepcopy(self.options[key])
            options[key]['value'] = data_dict[key]
            if options[key]['value'] is None:
                options[key]['value'] = ""

            # if vector store, set by idx, not by value
            if key == "vector_store":
                idx = self.get_choice_idx_by_id(data_dict[key])
                options[key]['idx'] = idx  # set by idx

        self.window.controller.config.load_options(self.id, options)  # set combo by idx

        # restore functions
        if assistant.has_functions():
            functions = assistant.get_functions()
            values = []
            for function in functions:
                values.append(
                    {
                        "name": function['name'],
                        "params": function['params'],
                        "desc": function['desc'],
                    }
                )
            self.window.ui.config[self.id]['tool.function'].items = values
            self.window.ui.config[self.id]['tool.function'].model.updateData(values)
        else:
            self.window.ui.config[self.id]['tool.function'].items = []
            self.window.ui.config[self.id]['tool.function'].model.updateData([])

        # set focus to name field
        self.window.ui.config[self.id]['name'].setFocus()

    def import_functions(self, force: bool = False):
        """
        Import functions from plugins

        :param force: force import
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='assistant.functions.import',
                id='',
                msg=trans('confirm.assistant.functions.import'),
            )
            return
        func_plugin = self.window.core.command.as_native_functions(True)
        values = self.window.ui.config[self.id]['tool.function'].items

        # import to editor
        for func in func_plugin:
            if func['name'] in [f['name'] for f in values]:
                # replace if exists
                for i in range(len(values)):
                    if values[i]['name'] == func['name']:
                        values[i] = {
                            "name": func['name'],
                            "params": func['params'],
                            "desc": func['desc'],
                        }
            else:
                # add new
                values.append(
                    {
                        "name": func['name'],
                        "params": func['params'],
                        "desc": func['desc'],
                    }
                )
        self.window.ui.config[self.id]['tool.function'].items = values
        self.window.ui.config[self.id]['tool.function'].model.updateData(values)

    def clear_functions(self, force: bool = False):
        """
        Clear functions list

        :param force: force clear
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='assistant.functions.clear',
                id='',
                msg=trans('confirm.assistant.functions.clear'),
            )
            return
        values = []
        self.window.ui.config[self.id]['tool.function'].items = values
        self.window.ui.config[self.id]['tool.function'].model.updateData(values)

    def save(self):
        """Save assistant"""
        created = False

        # get data from fields
        id = self.window.controller.config.get_value(
            parent_id=self.id, 
            key='id', 
            option=self.options['id'],
        )  # empty or not
        name = self.window.controller.config.get_value(
            parent_id=self.id, 
            key='name', 
            option=self.options['name'],
        )
        model = self.window.controller.config.get_value(
            parent_id=self.id, 
            key='model', 
            option=self.options['model'],
        )

        # check name
        if name is None or name == "" or model is None or model == "":
            self.window.ui.dialogs.alert(
                trans('assistant.form.empty.fields')
            )
            return

        if id is None or id == "" or not self.window.core.assistants.has(id):
            self.window.update_status(trans('status.sending'))
            QApplication.processEvents()
            assistant = self.window.controller.assistant.create()  # id is created in API here
            if assistant is None:
                self.window.update_status("status.error")
                print("ERROR: Assistant not created!")
                return
            id = assistant.id  # set to ID created in API
            self.window.core.assistants.add(assistant)
            self.window.controller.config.apply_value(
                parent_id=self.id,
                key="id",
                option=self.options["id"],
                value=id,
            )
            created = True
        else:
            assistant = self.window.core.assistants.get_by_id(id)

        # assign data from fields to assistant object
        self.assign_data(assistant)

        # update data in API if only updating data here (not creating)
        if not created:
            self.window.update_status(trans('status.sending'))
            QApplication.processEvents()
            self.window.controller.assistant.update_data(assistant)

        # save
        self.window.core.assistants.save()
        self.window.controller.assistant.refresh()
        self.window.controller.assistant.update()

        self.window.ui.dialogs.close('editor.assistants')
        self.window.update_status(trans('status.assistant.saved'))

        # switch to edited assistant
        self.window.controller.assistant.select_by_id(id)

    def assign_data(self, assistant: AssistantItem):
        """
        Assign data from fields to assistant

        :param assistant: assistant item
        """
        model = self.window.controller.config.get_value(
            parent_id=self.id,
            key='model',
            option=self.options['model'],
        )
        if model == '_':
            model = None

        assistant.name = self.window.controller.config.get_value(
            parent_id=self.id,
            key='name',
            option=self.options['name'],
        )
        assistant.model = model
        assistant.description = self.window.controller.config.get_value(
            parent_id=self.id,
            key='description',
            option=self.options['description'],
        )
        assistant.instructions = self.window.controller.config.get_value(
            parent_id=self.id,
            key='instructions',
            option=self.options['instructions'],
        )
        assistant.tools = {
            'code_interpreter': self.window.controller.config.get_value(
                parent_id=self.id,
                key='tool.code_interpreter',
                option=self.options['tool.code_interpreter'],
            ),
            'file_search': self.window.controller.config.get_value(
                parent_id=self.id,
                key='tool.file_search',
                option=self.options['tool.file_search'],
            ),
            'function': [],  # functions are assigned separately (below)
        }

        # assign assistant's functions tool
        values = self.window.controller.config.get_value(
                parent_id=self.id,
                key='tool.function',
                option=self.options['tool.function'],
        )
        functions = []
        for function in values:
            name = function['name']
            params = function['params']
            desc = function['desc']
            if name is None or name == "":
                continue
            if params is None or params == "":
                params = '{"type": "object", "properties": {}}'  # default empty JSON params
            if desc is None:
                desc = ""
            functions.append(
                {
                    "name": name,
                    "params": params,
                    "desc": desc,
                }
            )

        if len(functions) > 0:
            assistant.tools['function'] = functions
        else:
            assistant.tools['function'] = []

        # vector store
        store_id = self.get_selected_store_id()
        assistant.vector_store = store_id



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import hashlib
from typing import Optional, Tuple, List

from llama_index.core.indices.service_context import ServiceContext
from llama_index.core.indices.base import BaseIndex
from llama_index.core.indices.vector_store.base import VectorStoreIndex

from .base import BaseStore
from .ctx_attachment import CtxAttachmentProvider
from .temp import TempProvider


class Storage:
    def __init__(self, window=None):
        """
        Storage handler

        :param window: Window instance
        """
        self.window = window
        self.storages = {}
        self.indexes = {}
        self.tmp_storage = TempProvider(window=window)

    def get_storage(self) -> Optional[BaseStore]:
        """
        Get current vector store provider

        :return: vector store provider instance
        """
        current = self.window.core.config.get("llama.idx.storage")
        if current is None \
                or current == "_" \
                or current not in self.storages:
            return None
        return self.storages[current]

    def get_tmp_storage(self) -> Optional[TempProvider]:
        """
        Get temp vector store provider

        :return: vector store provider instance
        """
        return self.tmp_storage

    def get_ctx_idx_storage(self, path: str) -> CtxAttachmentProvider:
        """
        Get temp vector store provider
        
        :param path: Path to index on disk
        :return: vector store provider instance
        """
        return CtxAttachmentProvider(
            window=self.window,
            path=path
        )

    def register(self, name: str, storage: BaseStore):
        """
        Register vector store provider

        :param name: vector store provider name (ID)
        :param storage: vector store provider instance
        """
        storage.attach(window=self.window)
        self.storages[name] = storage

    def get_ids(self) -> List[str]:
        """
        Return all vector store providers IDs

        :return: list of vector store providers (IDs)
        """
        return list(self.storages.keys())

    def exists(self, id: Optional[str] = None) -> bool:
        """
        Check if index exists

        :param id: index name
        :return: True if exists
        """
        storage = self.get_storage()
        if storage is None:
            raise Exception('Storage engine not found!')
        return storage.exists(id)

    def create(self, id: str):
        """
        Create empty index

        :param id: index name
        """
        storage = self.get_storage()
        if storage is None:
            raise Exception('Storage engine not found!')
        storage.create(id)

    def get(
            self,
            id: str,
            service_context: Optional[ServiceContext]
    ) -> BaseIndex:
        """
        Get index instance

        :param id: index name
        :param service_context: service context
        :return: index instance
        """
        storage = self.get_storage()
        if storage is None:
            raise Exception('Storage engine not found!')
        return storage.get(
            id=id,
            service_context=service_context,
        )

    def store(
            self,
            id: str,
            index: Optional[BaseIndex] = None
    ):
        """
        Store index

        :param id: index name
        :param index: index instance
        """
        storage = self.get_storage()
        if storage is None:
            raise Exception('Storage engine not found!')
        storage.store(
            id=id,
            index=index,
        )

    def remove(self, id: str) -> bool:
        """
        Clear index only

        :param id: index name
        :return: True if success
        """
        storage = self.get_storage()
        if storage is None:
            raise Exception('Storage engine not found!')
        return storage.remove(id)

    def truncate(self, id: str) -> bool:
        """
        Truncate and clear index

        :param id: index name
        :return: True if success
        """
        storage = self.get_storage()
        if storage is None:
            raise Exception('Storage engine not found!')
        return storage.truncate(id)

    def remove_document(self, id: str, doc_id: str) -> bool:
        """
        Remove document from index

        :param id: index name
        :param doc_id: document ID
        :return: True if success
        """
        storage = self.get_storage()
        if storage is None:
            raise Exception('Storage engine not found!')
        return storage.remove_document(
            id=id,
            doc_id=doc_id,
        )

    def get_tmp(
            self,
            identifier: str,
            service_context: Optional[ServiceContext] = None
    ) -> Tuple[str, BaseIndex]:
        """
        Get tmp index instance

        :param identifier: identifier
        :param service_context: service context
        :return: index instance
        """
        # convert path to md5 hash
        id = hashlib.md5(identifier.encode()).hexdigest()
        storage = self.get_tmp_storage()
        if storage is None:
            raise Exception('Storage engine not found!')
        return id, storage.get(
            id=id,
            service_context=service_context,
        )

    def store_tmp(
            self,
            id: str,
            index: Optional[BaseIndex] = None
    ):
        """
        Store index

        :param id: index name
        :param index: index instance
        """
        storage = self.get_tmp_storage()
        if storage is None:
            raise Exception('Storage engine not found!')
        storage.store(
            id=id,
            index=index,
        )

    def count_tmp(self) -> int:
        """
        Count temp indices

        :return: number of temp indices
        """
        storage = self.get_tmp_storage()
        if storage is None:
            raise Exception('Storage engine not found!')
        return storage.count()

    def clean_tmp(self, id: str):
        """
        Clean temp index

        :param id: index name
        """
        storage = self.get_tmp_storage()
        if storage is None:
            raise Exception('Storage engine not found!')
        storage.clean(id)

    def get_ctx_idx(
            self,
            path: str,
            service_context: Optional[ServiceContext] = None
    ) -> BaseIndex:
        """
        Get context index instance

        :param path: path to index directory
        :param service_context: service context
        :return: index instance
        """
        # convert path to md5 hash
        storage = self.get_ctx_idx_storage(path)
        if storage is None:
            raise Exception('Storage engine not found!')
        return storage.get(
            id="",
            service_context=service_context,
        )

    def store_ctx_idx(
            self,
            path: str,
            index: Optional[BaseIndex] = None
    ):
        """
        Store context index

        :param path: path to index directory
        :param index: index instance
        """
        storage = self.get_ctx_idx_storage(path)
        if storage is None:
            raise Exception('Storage engine not found!')
        storage.store(
            id="",
            index=index,
        )

    def clean_ctx_idx(self, path: str):
        """
        Clean temp index

        :param path: path to index directory
        """
        storage = self.get_ctx_idx_storage(path)
        if storage is None:
            raise Exception('Storage engine not found!')
        storage.clean()

    def index_from_empty(self) -> BaseIndex:
        """
        Create empty index

        :return: index instance
        """
        return VectorStoreIndex([])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.26 02:00:00                  #
# ================================================== #

import json


class AttachmentItem:

    TYPE_FILE = 'file'
    TYPE_URL = 'url'

    def __init__(self):
        """
        Attachment item
        """
        self.name = None
        self.id = None
        self.uuid = None
        self.path = None
        self.remote = None
        self.vector_store_ids = []
        self.meta_id = None
        self.ctx = False
        self.consumed = False
        self.size = 0
        self.send = False
        self.type = self.TYPE_FILE

    def serialize(self) -> dict:
        """
        Serialize item to dict

        :return: serialized item
        """
        return {
            'id': self.id,
            'uuid': self.uuid,
            'name': self.name,
            'path': self.path,
            'size': self.size,
            'remote': self.remote,
            'ctx': self.ctx,
            'vector_store_ids': self.vector_store_ids,
            'type': self.type,
            'meta_id': self.meta_id,
            'send': self.send
        }

    def deserialize(self, data: dict):
        """
        Deserialize item from dict

        :param data: serialized item
        """
        if 'id' in data:
            self.id = data['id']
        if 'uuid' in data:
            self.uuid = data['uuid']
        if 'name' in data:
            self.name = data['name']
        if 'path' in data:
            self.path = data['path']
        if 'size' in data:
            self.size = data['size']
        if 'remote_id' in data:
            self.remote = data['remote']
        if 'ctx' in data:
            self.ctx = data['ctx']
        if 'vector_store_ids' in data:
            self.vector_store_ids = data['vector_store_ids']
        if 'type' in data:
            self.type = data['type']
        if 'meta_id' in data:
            self.meta_id = data['meta_id']
        if 'send' in data:
            self.send = data['send']

    def dump(self) -> str:
        """
        Dump item to string

        :return: serialized item
        """
        try:
            return json.dumps(self.serialize())
        except Exception as e:
            pass
        return ""

    def __str__(self):
        """To string"""
        return self.dump()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.09 23:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QTabWidget

from pygpt_net.core.tabs.tab import Tab


class TabBody(QTabWidget):
    def __init__(self):
        super(TabBody, self).__init__()
        self.owner = None

    def setOwner(self, owner: Tab):
        """
        Set tab parent (owner)

        :param owner: parent tab instance
        """
        self.owner = owner

    def getOwner(self) -> Tab:
        """
        Get tab parent (owner)

        :return: parent tab instance
        """
        return self.owner


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.26 19:00:00                  #
# ================================================== #

import time

from PySide6.QtCore import Slot, Signal
from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    playback = Signal(object, str)
    stop = Signal()


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.text = None
        self.event = None
        self.cache_file = None  # path to cache file
        self.mode = "generate"  # generate|playback
        self.audio_file = None

    @Slot()
    def run(self):
        try:
            if self.mode == "generate":
                self.generate()
            elif self.mode == "playback":
                self.play()
        except Exception as e:
            self.error(e)

    def generate(self):
        """
        Generate and play audio file
        """
        if self.text is None or self.text == "":
            time.sleep(0.2)  # wait
            return
        path = self.plugin.get_provider().speech(self.text)
        if path:
            from pygame import mixer
            mixer.init()
            playback = mixer.Sound(path)
            self.stop_playback()  # stop previous playback
            playback.play()
            self.send(playback)  # send playback object to main thread to allow force stop

            # store in cache if enabled
            if self.cache_file:
                self.cache_audio_file(path, self.cache_file)

    def play(self):
        """
        Play audio file only
        """
        if self.audio_file:
            from pygame import mixer
            mixer.init()
            playback = mixer.Sound(self.audio_file)
            playback.play()
            self.send(playback)  # send playback object to main thread to allow force stop

    def cache_audio_file(self, src: str, dst: str):
        """
        Store audio file in cache

        :param src: source path
        :param dst: destination path
        """
        import shutil
        try:
            shutil.copy(src, dst)
            # print("Cached audio file:", dst)
        except Exception as e:
            self.error(e)

    def send(self, playback):
        """
        Send playback object to main thread

        :param playback: playback object
        """
        self.signals.playback.emit(playback, self.event)

    def stop_playback(self):
        """Stop audio playback"""
        self.stop()

    def stop(self):
        """Send stop signal to main thread"""
        self.signals.stop.emit()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional, List

from pygpt_net.core.types import (
    MODE_COMPLETION,
)
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem


class Completion:
    def __init__(self, window=None):
        """
        Completion wrapper

        :param window: Window instance
        """
        self.window = window
        self.input_tokens = 0

    def send(
            self,
            prompt: str,
            system_prompt: str,
            model: ModelItem,
            history: Optional[List[CtxItem]] = None,
            stream: bool = False,
            ai_name: Optional[str] = None,
            user_name: Optional[str] = None
    ):
        """
        Chat with LLM

        :param prompt: user prompt
        :param system_prompt: system prompt
        :param model: model item
        :param history: history
        :param stream: stream mode
        :param ai_name: AI name
        :param user_name: username
        :return: LLM response
        """
        llm = None
        if 'provider' in model.langchain:
            provider = model.langchain['provider']
            if provider in self.window.core.llm.llms:
                try:
                    # init
                    self.window.core.llm.llms[provider].init(
                        self.window,
                        model,
                        "langchain",
                        MODE_COMPLETION,
                    )
                    # get LLM provider instance
                    llm = self.window.core.llm.llms[provider].completion(
                        self.window,
                        model,
                        stream,
                    )
                except Exception as e:
                    self.window.core.debug.log(e)
                    raise e
        if llm is None:
            raise Exception("Invalid LLM")

        message = self.build(
            prompt=prompt,
            system_prompt=system_prompt,
            model=model,
            ai_name=ai_name,
            user_name=user_name,
        )
        if stream:
            return llm.stream(message)
        else:
            return llm.invoke(message)

    def build(
            self,
            prompt: str,
            system_prompt: str,
            model: ModelItem,
            history: Optional[List[CtxItem]] = None,
            ai_name: Optional[str] = None,
            user_name: Optional[str] = None
    ) -> str:
        """
        Build completion string

        :param prompt: user prompt
        :param system_prompt: system prompt
        :param model: model item
        :param history: history
        :param ai_name: AI name
        :param user_name: username
        :return: message string (parsed with context)
        """
        message = ""

        # tokens config
        used_tokens = self.window.core.tokens.from_user(
            prompt,
            system_prompt,
        )
        max_ctx_tokens = self.window.core.config.get('max_total_tokens')

        # fit to max model ctx tokens
        if max_ctx_tokens > model.ctx:
            max_ctx_tokens = model.ctx

        # input tokens: reset
        self.reset_tokens()

        if system_prompt is not None and system_prompt != "":
            message += system_prompt

        if self.window.core.config.get('use_context'):
            items = self.window.core.ctx.get_history(
                history,
                model.id,
                "langchain",
                used_tokens,
                max_ctx_tokens,
            )
            for item in items:
                if item.input_name is not None \
                        and item.output_name is not None \
                        and item.input_name != "" \
                        and item.output_name != "":
                    if item.final_input is not None and item.final_input != "":
                        message += "\n" + item.input_name + ": " + item.final_input
                    if item.final_output is not None and item.final_output != "":
                        message += "\n" + item.output_name + ": " + item.final_output
                else:
                    if item.final_input is not None and item.final_input != "":
                        message += "\n" + item.final_input
                    if item.final_output is not None and item.final_output != "":
                        message += "\n" + item.final_output

        # append names
        if user_name is not None \
                and ai_name is not None \
                and user_name != "" \
                and ai_name != "":
            message += "\n" + user_name + ": " + str(prompt)
            message += "\n" + ai_name + ":"
        else:
            message += "\n" + str(prompt)

        # input tokens: update
        self.input_tokens += self.window.core.tokens.from_text(
            message,
            model.id,
        )
        return message

    def reset_tokens(self):
        """Reset input tokens counter"""
        self.input_tokens = 0

    def get_used_tokens(self) -> int:
        """
        Get input tokens counter

        :return: input tokens counter
        """
        return self.input_tokens



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.09 03:00:00                  #
# ================================================== #

import os

from PySide6.QtCore import QSize
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QVBoxLayout, QLabel, QPushButton, QWidget, QSizePolicy, QHBoxLayout

from pygpt_net.ui.widget.textarea.name import NameInput
from pygpt_net.ui.widget.option.slider import OptionSlider
from pygpt_net.ui.widget.audio.input_button import VoiceControlButton
from pygpt_net.utils import trans

from .agent import Agent
from .agent_llama import AgentLlama
from .image import Image
from .indexes import Indexes
from .vision import Vision
from ...widget.option.toggle_label import ToggleLabel


class Footer:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window
        self.agent = Agent(window)
        self.agent_llama = AgentLlama(window)
        self.image = Image(window)
        self.indexes = Indexes(window)
        self.vision = Vision(window)

    def setup(self) -> QWidget:
        """
        Setup footer

        :return: QHBoxLayout
        """
        # AI and users names
        # names_layout = QHBoxLayout()
        # names_layout.addLayout(self.setup_name_input('preset.ai_name', trans("toolbox.name.ai")))
        # names_layout.addLayout(self.setup_name_input('preset.user_name', trans("toolbox.name.user")))

        # bottom
        option = dict(self.window.controller.settings.editor.get_options()["temperature"])
        self.window.ui.nodes['temperature.label'] = QLabel(trans("toolbox.temperature.label"))
        self.window.ui.config['global']['current_temperature'] = \
            OptionSlider(self.window, 'global', 'current_temperature', option)
        self.window.ui.add_hook("update.global.current_temperature", self.window.controller.mode.hook_global_temperature)

        # voice control btn
        self.window.ui.nodes['voice.control.btn'] = VoiceControlButton(self.window)
        self.window.ui.nodes['voice.control.btn'].setVisible(False)

        # per mode options
        rows = QVBoxLayout()
        # rows.addWidget(self.window.ui.nodes['temperature.label'])
        # rows.addWidget(self.window.ui.config['global']['current_temperature'])
        rows.addWidget(self.agent.setup())
        rows.addWidget(self.agent_llama.setup())
        rows.addWidget(self.image.setup())
        # rows.addWidget(self.vision.setup())
        rows.addWidget(self.indexes.setup_options())
        rows.addWidget(self.window.ui.nodes['voice.control.btn'])
        rows.setContentsMargins(2, 0, 0, 0)

        self.window.ui.nodes['layout.split'] = ToggleLabel(trans('layout.split'), label_position="left",
                                                          icon=":/icons/split_screen.svg")
        self.window.ui.nodes['layout.split'].box.stateChanged.connect(
            lambda: self.window.controller.ui.tabs.toggle_split_screen(self.window.ui.nodes['layout.split'].box.isChecked())
        )
        split_layout = QHBoxLayout()

        split_layout.addWidget(QLabel(""))
        split_layout.addStretch(1)
        split_layout.addWidget(self.window.ui.nodes['layout.split'])
        split_layout.setContentsMargins(5, 0, 15, 0)
        split_widget = QWidget()
        split_widget.setLayout(split_layout)
        rows.addWidget(split_widget)

        # logo
        logo_button = self.setup_logo()

        # bottom (options and logo)
        # bottom = QHBoxLayout()
        # bottom.addLayout(rows, 80)
        # bottom.addWidget(logo_button, 20)
        # bottom.setStretchFactor(logo_button, 1)
        # bottom.setAlignment(logo_button, Qt.AlignRight | Qt.AlignBottom)
        # bottom_widget = QWidget()
        # bottom_widget.setLayout(bottom)

        # layout rows
        # layout = QVBoxLayout()
        # layout.addLayout(names_layout)
        # layout.addWidget(bottom_widget)

        widget = QWidget()
        widget.setLayout(rows)

        return widget

    def setup_name_input(self, id: str, title: str) -> QVBoxLayout:
        """
        Setup name input

        :param id: ID of the input
        :param title: Title of the input
        :return: QVBoxLayout
        """
        label_key = 'toolbox.' + id + '.label'
        self.window.ui.nodes[label_key] = QLabel(title)
        self.window.ui.nodes[id] = NameInput(self.window, id)

        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.nodes[label_key])
        layout.addWidget(self.window.ui.nodes[id])

        return layout

    def setup_logo(self) -> QPushButton:
        """
        Setup logo

        :return: QPushButton
        """
        path = os.path.abspath(os.path.join(self.window.core.config.get_app_path(), 'data', 'logo.png'))

        button = QPushButton()
        button.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        button.setIcon(QIcon(path))
        button.setIconSize(QSize(100, 28))
        button.setFlat(True)
        button.clicked.connect(lambda: self.window.controller.dialogs.info.goto_website())

        return button
